;; Generated package description from mode: lisp-data; .el  -*- consultno-byte-compile: t -*-
(define-package "consult" "2.2" "Consulting completing-read" '((emacs "28.1") (compat "30")) :commit "30a42ac8f3d42f653eda234ee538c1960704f4f2" :maintainer '("Daniel Mendler" . "mail@daniel-mendler.de") :keywords '("matching" "files" "completion") :url "https://github.com/minad/consult")
